;; Starting up:

;; \scheme{abort : TST -> void}
;; note: this variable is set upon start-up; for self-protection, its initial value is error
(define (abort x) (error 'abort "undefined"))
  
;; \scheme{walk : -> void}
;; effect: set abort and start the read-eval-print shell
(define (walk)
  [(call-with-current-continuation
     (lambda (k) (set! abort k)
       (set-pc! 0)
       (shell)))])


;; The interactive command shell: 

;; \scheme{shell: -> void}
;; note: the function implements a read-command-eval-loop
(define shell
  (local ()
    (lambda ()
      (print-cell (get-pc))
      (let ([command (prompt-read " % ")])
	(cond
	  [(cons? command) (install (assemble command))]
	  [(and (number? command) (value? command)) (install command)]
	  [(symbol? command) ((find-command command))]
          [(eof-object? command) 
           (if (empty? input-stack) 
               (top-level-quit)  ; top-level ^D means 'q.
               (begin ; Pop back from 'load command:
                      (current-input-port (first input-stack)) 
                      (set! input-stack (rest input-stack))))]
	  [else (default-action command)]))
      (shell))))

(define input-stack empty)


;;
(define (load-file)
    (local [(define file-name 
              (prompt-read "Name of script file (if case-sensitive, put in quotes): "))
            (define file-string
              (if (symbol? file-name) (symbol->string file-name) file-name))
            (define file
              (with-handlers 
                 [(exn? (lambda (exn) (printf "~a~n" (exn-message exn))
                                      (printf "load-file ignored.~n" )))]
                  (open-input-file file-string)))
             ]
        (if (not (void? file))
            (begin (set! input-stack (cons (current-input-port) input-stack))
                   (current-input-port file)))))






;; \scheme{install : value -> address}
;; effect: store value in mem[PC], print as command, increment PC
(define (install value)
  (set-mem! (get-pc) value)
  (print-cell (get-pc))
  (newline)
  (get&inc-pc))

;\newpage
;; Some additional shell commands

;; \scheme{step : -> void}
(define (step)
  (let ((the-value (get-mem (get&inc-pc))))
    (printf "~n~s~n" (disassemble the-value))
    (execute (sign the-value) (decode the-value))
    (print-regs)))

;; \scheme{reset! : -> void}
;; effect: set all memory cells, registers, and the PC to 0
;; breaks abstraction!
(define (reset!)
  (vector-fill! mem-array 0)
  (vector-fill! reg-array 0)
  (set-pc! 0))

;; \scheme{show : -> void}
;; effect: print contents of a range of memory cells, interactive
(define (show)
  (let* {[low (prompt-read "start: ")]
	 [high (prompt-read "end: ")]}
    (if (and (number? low) (address? low)
	     (number? high) (address? high)
  	     (<= low high))
	(let S ((low low))
	  (print-cell low) (newline)
	  (if (= low high) (newline) (S (add1 low))))
	(begin
	  (printf "Please enter two numbers, n and m, such that n <= m.~n")
	  (show)))))

;; \scheme{print-cell : address -> void}
;; effect: print content of memory cell 
(define (print-cell addr)
  (let ([val (get-mem addr)])
    (printf "mem[~a]: ~a ~s"
      (format-number addr DIGITS-IN-ADD "")
      (format-number val  DIGITS        (if (negative? val) "-" "+"))
      (disassemble val))))

;; \scheme{print-regs : -> void}
;; effect: print contents of registers
(define (print-regs)
  (let ((reg (box 0)))
    (printf "PC:~s" (get-pc))
    (for-each (lambda (val)
		(printf " R~s:~s" (unbox reg) val)
		(set-box! reg (add1 (unbox reg))))
      (vector->list reg-array))
    (printf "~n")))

;; Input and output auxiliaries

;; \scheme{format-number : num num str -> str}
(define (format-number n width prefix)
  (let* ((n-as-string (number->string (abs n)))
	 (spaces-of-n (string-length n-as-string))
	 (no-of-lders (- width spaces-of-n)))
    (string-append
      prefix
      (make-string no-of-lders #\0)
      n-as-string)))

;; \scheme{prompt-read : str -> TST}
;; effect: print string and return next S-expression read from standard input
;;
;; BUG?: If char-ready? returned whether or not
;;  "read" will have to block, we'd be okay.
;;  But char-ready? seems to always return true after the first call.
;;
(define (prompt-read prmpt)
  (let* {[had-to-wait? (not (char-ready?))]
         ;[_ (printf "  <had-to-wait? = ~s>  " had-to-wait?)]
         [_ (printf prmpt)]
         [answer (read)]
         [_ (unless had-to-wait? (printf "~s~n" answer))]
         }
    answer))

;; \scheme{default-action : -> void}
;; effect: print help
(define (default-action unk-cmd)
  (unless (void? unk-cmd) (printf "~nUnknown command: ~s.~n" unk-cmd))
  (printf "  [instr] - set current cell to [instr]'s encoding; move to next cell.~n")
  (printf "  [num]   - set current cell to [num];              move to next cell.~n")
  (for-each (lambda (e) (printf "  ~s - ~a~n" (entry-name e)
                                              (entry-description e))) 
            the-command-list)
  )

;\newpage
;; The table of shell commands

;; structure definition for command table: lists so that assq works
(define-struct entry (name description action))


;; confirmed? : string --> boolean
;; Prompt for a yes/no comfirmation,
;; and print a consoling message if not.
;;
(define (confirmed? str)
  (let* {[affirmations
          '(y Y  yes Yes YES   yep Yep YEP yeah Yeah YEAH sure Sure SURE)]
         [reply (prompt-read (string-append str " [yn] "))]
         [confirmed (memq reply affirmations)]}
     (unless confirmed (printf "Never mind.~n"))
     confirmed))


;; top-level-quit:
;; exit the program.
;;
(define (top-level-quit)
  (abort (lambda () (printf "Done walking -- must rest.~n"))))

;; list of available shell commands, names, and comments
;; \scheme{the-command-list : (listof (structure:entry sym str (-> void)))}
(define the-command-list
  (list 
    (make-entry 'l "load a script containing further machine inspector commands" load-file)
    (make-entry 'r "reset" 
      (lambda () (when (confirmed? "Reset?") 'y) (reset!)))
    (make-entry 'm "show memory" show)
    (make-entry 's "show registers and pc" print-regs)

    ; Note that set-pc! will error if it gets a non-number as input.
    ; An older version just ignored the set-pc! if it weren't an
    ; address, w/o giving an error message, which is worse. --ian 01.nov
    ;
    (make-entry 'p "set pc" (lambda ()  (set-pc! (prompt-read "  new pc: "))))
    (make-entry '+ "increment pc" (lambda () (set-pc! (add1 (get-pc)))))
    (make-entry '- "decrement pc" (lambda () (set-pc! (sub1 (get-pc)))))

    (make-entry 'x "execute" 
      (lambda ()
	(when (confirmed? (format "Execute starting at ~s?" (get-pc)))
	  (printf "Executing ...~n")
	  (run)  
          ; (run) doesn't return -- it ends by 
          ; calling continuation "walk", instead ?
          ; So this line never gets executed, oh well:
	  (printf "~n... done executing.~n"))))

    (make-entry 'n "execute next instruction only (\"step\")" step)


    (make-entry '?    "show this help" (lambda () (default-action (void))))
    (make-entry 'h    "show this help" (lambda () (default-action (void))))
    (make-entry 'help "show this help" (lambda () (default-action (void))))

    (make-entry 'exit "quit" top-level-quit)
    (make-entry 'quit "quit" top-level-quit)
    (make-entry 'q    "quit" top-level-quit)
    ))

;; \scheme{find-command : sym -> (-> void)}
(define (find-command command)
  (let ((r (assf (lambda (x) (eq? (entry-name x) command)) the-command-list)))
    (if r (entry-action r) (lambda () (default-action command)))))
